 /*
*	Developed by Felix Hahn
*	This program performes a frequenzy domain convolution nearly in real-time 
*	with the input signal and the chosen filter coefficients.
*	It works interrupt-based. AIC_RX() is the ISR and is 
*	executed when AIC receives data. 
*/
/*
*	begin of program
*/
#define SIZE 1024			//size of buffer
#define LED 4				//number of LEDs
#define DIPS 4				//number of switches
#define LOOP_SIZE 3500		//number of max loop iterations
#define HOPSIZE 512			//define hopsize
//	define const for theater IR
#define	P	79				//number of colums (Theater)
//#define Q	40				//Q=A1
//#define R	41984			//R=A2
/*
*	include DSK libraries
*/
#include "dsk6713.h"			//for init DSK
#include "dsk6713_aic23.h"		//for init AIC23 codec
#include "dsk6713_dip.h"		//for init switches
#include "dsk6713_led.h"		//for init LEDs


/*
*	include coefficients
*/
#include "tiefpass.h"		//coefficients of lowpass
#include "bandpass.h"		//coefficients of bandpass
#include "bandstopp.h"		//coefficients of bandstop
#include "hochpass.h"//coefficients of highpass
#include "LR_short.h"		//coefficients of living room impulse response
#include "TH_IR_512.h"	//coefficients of theater impulse response
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#ifndef PI
# ifdef M_PI
#  define PI M_PI
# else
#  define PI 3.14159265358979323846
# endif
#endif
/*
*	function prototypes  
*/

void AIC_RX(void);			//ISR
void setLED(void);			//set LEDs
int readDips();				//read state of switches
int gen_twiddle(float *w, int n);	//generates Twiddlefactor
void bit_rev(float *xb, int nrn);		//bitreverse vector x
void divide(float *xd, int ndn);		//function to divide
void DSPF_sp_icfftr2_dif(float *xi, float *w, int nin);
void DSPF_sp_cfftr2_dit(float *xr, float *w, int nen);
void DSPF_q15tofl(short *xq, float *r, int nxq);
void DSPF_fltoq15(float *xf, short *rq, int nxf);
short T_BUFFER[SIZE * 2];		//input buffer
float OUT_BUFFER[(P+2) * HOPSIZE];		//output buffer
short OUT[SIZE];
float in[SIZE * 2];				//copy of input buffer
float in_Spektrum[SIZE * 2];	//spectrum of inputarray
float Teilfaltung_Spektrum[P][SIZE * 2];	//spectrum of block convolution
float IR[P][SIZE * 2];
float Twiddle[SIZE];			//Twiddlefacor	
float *COEF;			//pointer to coefficients array
int COEF_len;				//length of coefficients array
volatile short BUFFER_FULL = 0;		//signalflag
volatile short BUFFER_FINISH = 0;
volatile short pos_OUT = 0;			//pointer to circular buffer of OUTPUT
Int16 dip; 					//return variable for readDips();				
short yn;					//output variable
float y[(2+P) * HOPSIZE];
volatile int pos = 0;				//pointer for circular buffer
short n2, ie, ia, i, j, k, m, ief, iaf;                                    
float rtemp, itemp, c, s;				
float *x;
float *rq;
int l, v, p, nrn, n, nen, ndn, nxq, nxf, nin;
volatile short first_spectrum = 0;	//flag for programcontrol
float a[P][SIZE];				//odd blockconvolution array
float b[P][SIZE];				//even blockconvolution array
float A[(P+1)* HOPSIZE];			//odd blockconvolution vector
float B[(P+1)* HOPSIZE];			//even blockconvolution vector
Uint32 t=0;					//workingvariabel

union 						//structure for stereo input
	{
	Uint32 rawdata;
	short channel[2];		//two channel for right and left
	}temp;					//alias signifier	
/*
*	setup for codec
*/
DSK6713_AIC23_CodecHandle hCodec;
DSK6713_AIC23_Config config = 
{ 
    0x0017, /* Set-Up Reg 0       Left line input channel volume control */                                                                                                               
    0x0017, /* Set-Up Reg 1       Right line input channel volume control */                                                                                                             
    0x01f9, /* Set-Up Reg 2       Left channel headphone volume control */                                                                                          
    0x01f9, /* Set-Up Reg 3       Right channel headphone volume control */                                                                                    
    0x0011, /* Set-Up Reg 4       Analog audio path control */                                                                                                         
    0x0000, /* Set-Up Reg 5       Digital audio path control */                                                                                                  
    0x0000, /* Set-Up Reg 6       Power down control */                                                                                                                            
    0x0043, /* Set-Up Reg 7       Digital audio interface format */                                                                                                       
    0x0081, /* Set-Up Reg 8       Sample rate control */                                                                                                     
    0x0001  /* Set-Up Reg 9       Digital interface activation */                                           
};		
/*
*	read state of switches
*/					
int readDips()
{
	int total=0; 			//return variable
	int i; 					//working variable
	/*
	*	loop to read all four switches, add
	*	with bitshifting if switch is ON and
	*	return decimal value
	*/
	for(i=0;i<DIPS;i++)
	{
		total|=(((DSK6713_DIP_get(i)==0) ? 1 : 0)<<i);
	}
	return total;
}
/*
*	check state of swtiches and
*	set appropriate LED on or off
*/


void setLED(void)
{
	int i; 					//working variable
	for(i=0;i<LED;i++)
	{
		/*
		*	if swtich is on set LED on
		*/
		if(DSK6713_DIP_get(i)==0)
		{
			DSK6713_LED_on(i);
		}
		/*
		*	if swtich is off set LED off
		*/
		else
		{
			DSK6713_LED_off(i);
		}
	}
	return;
}
/*
*generate Twiddlefactor for FFT and IFFT
*function from TI
*/

int gen_twiddle(float *w, int n)
	{
    double delta = 2 * PI / n;
    int i;
    for(i = 0; i < (n/2); i++)
    {
        w[2 * i + 1] = sin(i * delta);
        w[2 * i] = cos(i * delta);
    }

	return n;
	}  

/*
*bit-reverse
*function from TI
*/
void bit_rev(float *xb, int nrn)
{
     int i, j, k;
     float rtemp, itemp;

     j = 0;
     for(i=1; i < (nrn-1); i++)
        {
        k = nrn >> 1;
        while(k <= j)
             {
             j -= k;
             k >>= 1;
             }
        j += k;
        if(i < j)
          {
           rtemp    = xb[j * 2];
           xb[j * 2]   = xb[i * 2];
           xb[i * 2]   = rtemp;
           itemp    = xb[j * 2+1];
           xb[j * 2+1] = xb[i * 2+1];
           xb[i * 2+1] = itemp;
          }
        }
		return;
}
/*
*divide-function
*function from TI
*/
void divide(float *xd, int ndn)
{
	int i;
	float inv = 1.0/ ndn;
	for(i=0;i < ndn; i++)
	{
		xd[2 * i] = inv * xd[2 * i];
		xd[2 * i+1] = inv * xd[2 * i+1];
	}
	return;
}
/*
FFT function from TI
*/
void DSPF_sp_cfftr2_dit(float *xf, float *w, int nen)                     
      {   
     	                                                                                      
         n2 = nen;                                                          
         ief = 1;                                                          
                                                                          
         for(k=nen; k > 1; k >>= 1)                                         
         {                                                                
            n2 >>= 1;                                                     
            iaf = 0;                                                       
            for(j=0; j < ief; j++)                                         
            {                                                             
               c = w[2 * j];                                                
               s = w[2 * j + 1];                                              
               for(i=0; i < n2; i++)                                      
               {                                                          
                  m = iaf + n2;                                                              
                  rtemp     = c * xf[2 * m]   + s * xf[2 * m+1];                
                  itemp     = c * xf[2 * m+1] - s * xf[2 * m];                  
                  xf[2 * m]    = xf[2 * iaf]   - rtemp;                          
                  xf[2 * m+1]  = xf[2 * iaf+1] - itemp;                          
                  xf[2 * iaf]   = xf[2 * iaf]   + rtemp;                          
                  xf[2 * iaf+1] = xf[2 * iaf+1] + itemp;                          
                  iaf++;                                                   
               }                                                          
               iaf += n2;                                                   
       		 }                                                             
            ief <<= 1;                                                     
        } 
        return;                                                               
      } 
/*
IFFT function from TI
*/
void DSPF_sp_icfftr2_dif(float *xi, float *w, int nin)                   
          {                                                                                                                        
              n2 = 1;                                                     
              ie = nin;                                                     
              for(k=nin; k > 1; k >>= 1)                                    
              {                                                           
                  ie >>= 1;                                               
                  ia = 0;                                                 
                  for(j=0; j < ie; j++)                                   
                  {                                                       
                      c = w[2 * j];                                         
                      s = w[2 * j+1];                                       
                      for(i=0; i < n2; i++)                               
                       {                                                   
                          m = ia + n2;                                    
                          rtemp     = xi[2 * ia]   - xi[2 * m];                 
                          xi[2 * ia]   = xi[2 * ia]   + xi[2 * m];                 
                          itemp     = xi[2 * ia+1] - xi[2 * m+1];               
                          xi[2 * ia+1] = xi[2 * ia+1] + xi[2 * m+1];               
                          xi[2 * m]    = c * rtemp   - s * itemp;                
                          xi[2 * m+1]  = c * itemp   + s * rtemp;                
                          ia++;                                           
                      }                                                   
                      ia += n2;                                           
                  }                                                       
                  n2 <<= 1;                                               
              }                                                           
        return;
           }    

void DSPF_q15tofl(short *xq, float *r, int nxq)                             
     {                                                                    
      int i;                                                              
      for (i=0;i<nxq;i++)                                                  
     	{
		
     	 r[i]=(float)xq[i]/0x8000;                                       
     	}
		return;
     }     

void DSPF_fltoq15(float *xf, short *rq, int nxf)                                                                    
     {                                                                    
         int i, j;                                                        
                                                                          
         for(i = 0; i < nxf; i++)                                          
         {                                                                
             j = floor(32768 * xf[i]);                                     
                                                                          
             // saturate to 16-bit //                                     
             if (j>32767)  j =  32767;                                    
             if (j<-32768) j = -32768;                                    
                                                                          
             rq[i] = (short) j;                                            
         }     
         return;                                                           
     }   
/*
*	ISR
*	save new input value from codec in buffer,
*	calculate output value and write to codec.
*	circular buffer is used
*/
void AIC_RX(void)
{
	

	/*
	*	read new input value from codec and safe on rawdata
	*/
	DSK6713_AIC23_read(hCodec, &temp.rawdata);
	/*
	*	save right channel value on circular buffer position pos
	*/
	T_BUFFER[pos] = temp.channel[0];
   	
	
	
		yn = OUT[pos_OUT];	//load output to outputvariabel
			
	/*
	*	else convolve input buffer with chosen coefficients	FFT_Algorithmus used
	*/
	DSK6713_AIC23_write(hCodec, (Uint32) yn);
	/*
	*	update buffer pointer and save Buffer if buffer is full
	*/
	if(pos == (SIZE-2))
	{
	pos=1;
	}
	else
		{
		if (pos == (SIZE-1))
			{
			DSPF_q15tofl(T_BUFFER, in, (SIZE));
			pos=0;				//back to start of buffer		
			BUFFER_FULL=1;
			BUFFER_FINISH=1;
            DSPF_fltoq15(OUT_BUFFER, OUT, 1024);
            pos_OUT=0;
		}
		else
			{
			pos=(pos+2);
			pos_OUT++;
            }
		}	
	
	/*
	*	update pointer to output...set flag (Buffer_finish) if HOPSIZE samples put out
	*/

	

	return;
}
/*
*	main function
*/
main()
{

	float *COEF = (float*) FFT_TH;	//Floating coeffeicients

	

	/*
	*	set space for buffers
	*/
	memset(T_BUFFER, 0, sizeof(short) * SIZE * 2);	//Zeropadding inputbuffer
	memset(in, 0, sizeof(float) * SIZE * 2);	//Zeropadding Input buffer buffer
	memset(in_Spektrum, 0, sizeof(float) * SIZE * 2);	//Prepare array for FFT
	memset(Teilfaltung_Spektrum, 0, sizeof(float) * SIZE * 2 * P);	//Prepare array for Blockconvolution
	memset(a, 0, sizeof(float) * SIZE * 2 * P);	//odd blockconvolution array
 	memset(b, 0, sizeof(float) * SIZE * 2 * P);	//even blockconvolution array
	memset(OUT, 0, sizeof(short) * SIZE);
	memset(OUT_BUFFER, 0, sizeof(float) * (P + 2) * HOPSIZE);
	memset(Twiddle, 0, sizeof(float) * SIZE);
	/*
	*	initialize DSK and start codec
	*/
	DSK6713_init();
	hCodec = DSK6713_AIC23_openCodec(0,&config);
	/*
	*	set samp le frequency to 16kHz
	*/
	DSK6713_AIC23_setFreq(hCodec, DSK6713_AIC23_FREQ_16KHZ);
	/*
	*generate twiddlefactors in w  [HOPSIZE elements]	/	bitreverse it
	*/		
	gen_twiddle(Twiddle, SIZE);
	bit_rev(Twiddle, HOPSIZE);
    /*
     *generate Spectra of IR
     */
    memset(IR, 0, sizeof(float) * SIZE * 2 * P); //Array for blocks of IR (RE+IM)
    
    for(p=0; p<=P; p++)
        {
            t=0;   
            l=p * HOPSIZE;
			while(t<SIZE)
			{		
			IR[p][t]=COEF[l];
            t=t+2;
			l++;
			}
            DSPF_sp_cfftr2_dit(IR[p],Twiddle, SIZE);
        }
		gen_twiddle(Twiddle, SIZE);
		bit_rev(Twiddle, HOPSIZE);
	/*
	*	enable interrupts
	*/
	IRQ_enable(IRQ_EVT_RINT1);
	IRQ_globalEnable();
	/*
	*	active waiting for interrupt
	*	while no interrupt exists, do algorithmus from Zlzer
	*	write outputvariable y  
	*/
	
	while(1) 
	{
		
	dip=readDips();
	setLED();
	/*
	*waitingloop for full Buffer
	*/
	

	while(BUFFER_FULL == 0)
	{
	; 
	 }
    BUFFER_FULL = 0;
	/*
	*copy buffer
	*/

 	for(t=0; t<=((2 * SIZE)-1);t++)
			{
			in_Spektrum[t]=in[t];
			}
	
	/*
	*compute FFT ->output in bit-reversed order
	*/
	DSPF_sp_cfftr2_dit(in_Spektrum, Twiddle, 1024);	//compute FFT
	//bit_rev(in_Spektrum, (SIZE));						//bit-reverse output
	/*
	*multiplicate spectra elementwise for each partial block then IFFT, scale and sort
	*/
	
 	for(t=0;t < P ;t++)
	{
 		for(p=0; p < (2 * SIZE)-2; p=p+2)
			{
			Teilfaltung_Spektrum[t][p] = (in_Spektrum[p] * IR[t][p]) - (in_Spektrum[p+1] * IR[t][p+1]);
			Teilfaltung_Spektrum[t][p+1] = (in_Spektrum[p+1] * IR[t][p]) + (in_Spektrum[p] * IR[t][p+1]);
			}
        DSPF_sp_icfftr2_dif(Teilfaltung_Spektrum[t], Twiddle,  1024);
        divide(Teilfaltung_Spektrum[t],  SIZE);
		
		for(p=0; p <= (SIZE - 1);p++)
			{
			a[t][p]=Teilfaltung_Spektrum[t][2 * p] * 0.125;
			b[t][p]=Teilfaltung_Spektrum[t][(2 * p) +1] * 0.125;
			}
	}
	/*
	*Bestimmung der Teilfaltung
	*Overlapp add a and b for outputvector
	*/
	memset(A, 0, sizeof(float) * (P+1) * HOPSIZE);	//erase odd blockconvolution vector
	memset(B, 0, sizeof(float) * (P+1) * HOPSIZE);	//erase even blockconvolution vector

    t=0;
	l=0;
	p=0;
	while(  t<=(HOPSIZE*(P+1)))
		{	
		if(p>(SIZE-1))
			{
			t=t-HOPSIZE;
			l++;
			p=0;
			}
		if(l<=P)
			{
			A[t]=A[t]+a[l][p];
			B[t]=B[t]+b[l][p];
			}
		t++;
		p++;
		}
		
	
	/*
	*Add A(k) and B(k+hopsize)
	*/


	memset(y, 0, sizeof(float) * (P + 2) * HOPSIZE);				
p=0;
	 for(p=0; p <= (((P+2) * HOPSIZE) -1); p++)
	  	{
		if(p>((P+1) * HOPSIZE )-1)
			{
			y[p]=B[p-HOPSIZE];
			}
		else
			{
	 		if((int)(p-HOPSIZE)<0)
				{
				y[p]=y[p]+A[p];
				}
			else
				{
				y[p]=y[p]+A[p]+B[p-HOPSIZE];
				}
			}
		}
	 	/*
		*write outputbuffer  
		*/
	
	 if(first_spectrum == 0)
		{
		for(p=0; p <= ((P+2) * HOPSIZE -1); p++)
			{
			OUT_BUFFER[p]=y[p];
			}
		first_spectrum=1;

		}
		else
			{
	
			for(p=0; p <=  (P+2) * HOPSIZE -1; p++)
				{
				if(p <= (P * HOPSIZE - 1))
					{
					OUT_BUFFER[p]=OUT_BUFFER[p + (2 * HOPSIZE)] +y[p];
					}
				else
					{
					OUT_BUFFER[p]=y[p];
					}
				}
	
			 
	 		} //end of else
			
	 	} //end of loop 	
		  
} // end of mainfunction 



